/*
 * $Id: arch-solaris.c,v 1.1.1.1.2.1 1997/03/04 03:10:50 fengmd Exp $
 */
/*
 * $Log: arch-solaris.c,v $
 * Revision 1.1.1.1.2.1  1997/03/04 03:10:50  fengmd
 * Merge the change of the Cilk-5 base.
 *
 * Revision 1.3  1997/03/02 21:06:26  randall
 * Made sure Cilk_die doesn't return.
 *
 * Revision 1.2  1997/03/01 17:42:34  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdarg.h>
#include <stdio.h>
#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/tspriocntl.h>
#include <sys/procfs.h>

DECLARE_SHARED_RO_VARIABLE(static int, lock_type, );
#define MEMORY_LOCKS 0
#define OS_LOCKS 1

/***********************************************************\
 * Debugging & tracing
\***********************************************************/

DECLARE_SHARED_VARIABLE(static Cilk_mutex, dprintf_lock, );

void Cilk_dprintf(const char *fmt,...)
{
     va_list l;

     /* ensure nice interleaving of debugging messages */
     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: ", Self);
     else
	  fprintf(stderr, "global: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     fflush(stderr);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));
}

DECLARE_SHARED_VARIABLE(static Cilk_mutex, die_lock, );

void Cilk_die(const char *fmt,...)
{
     va_list l;

     Cilk_mutex_wait(&USE_SHARED(dprintf_lock));
     if (Self >= 0)
	  fprintf(stderr, "%d: fatal error: ", Self);
     else
	  fprintf(stderr, "global: fatal error: ");
     va_start(l, fmt);
     vfprintf(stderr, fmt, l);
     va_end(l);
     Cilk_mutex_signal(&USE_SHARED(dprintf_lock));

     /* ensure only the first fatal error produces death of the program */
     if (Cilk_mutex_try(&USE_SHARED(die_lock)))
	  abort();		/* dump core */
     else
          exit(0);              /* don't dump core */
}

/***********************************************************\
 * Mapping of variable segments
\***********************************************************/

#if !CILK_MANAGE_SHARED_SEGMENT 
/*
 * if linker segments are not used, a portable version
 * of these two functions is provided in cilk.c
 */

/* 
 * map the global variables segment as shared
 */
void Cilk_map_shared_segment(void)
{
     extern _end_Cilk_shared;	/* linker_defined symbols */
     extern _begin_Cilk_shared;
     char name[1024];
     int fd;
     void *startmem, *endmem;

     strcpy(name, "/tmp/_Cilk_gvars_XXXXXX");
     mktemp(name);

     startmem = &_begin_Cilk_shared;
     endmem = &_end_Cilk_shared;

     /* write all shared data to the file */
     fd = open(name, O_RDWR | O_CREAT, 0600);
     CILK_CHECK(fd >= 0, ("Open error\n"));
     unlink(name);
     CILK_CHECK(write(fd, startmem, endmem - startmem) == endmem - startmem,
		("error in write\n"));

     /* map shared data to be shared */
     if (startmem != mmap(startmem, endmem - startmem, PROT_READ | PROT_WRITE,
			  MAP_SHARED | MAP_FIXED, fd, 0))
	  Cilk_die("couldn't map shared part to file.\n");
     
     close(fd);
}
#endif

#if !CILK_MANAGE_PRIVATE_SEGMENT 
void Cilk_map_private_segment(void)
{
     /* variables are private by default */
}
#endif

/***********************************************************\
 * Malloc support
\***********************************************************/

/* Get a chunk of memory of size <len> that is mapped shared. */
void *Cilk_map_shared(size_t len)
{
     int fd;
     void *area;

     fd = open("/dev/zero", O_RDWR);

     CILK_CHECK(fd >= 0, ("Open error in Cilk_map_shared\n"));

     area = mmap(0, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
     if (area == MAP_FAILED) {
	  /* 
	   * can't use Cilk_die, because the mutex in
	   * not initialized yet
	   */
	  fprintf(stderr, "mmap error in Cilk_map_shared -- "
		   "couldn't allocate %d bytes\n", len);
	  exit(1);
     }
     close(fd);
     return area;
}

/***********************************************************\
 * Architecture-specific options
\***********************************************************/

/*
 * Parse any architecture-specific command-line options.  <argc>
 * and <argv> are the original arguments passed on the command line,
 * and <i> is the index of the first architecture-specific option (if any).
 * This function returns the index of the first user option, or <argc> if
 * there aren't any.
 */
int Cilk_arch_specific_options(int i, int argc, char *argv[])
{
     USE_SHARED_RO(lock_type) = MEMORY_LOCKS;

     for (; i < argc; ++i) {
	  if (*argv[i] != '-')
	       break;

	  if (strcmp(argv[i] + 1, "-") == 0) {
	       ++i;
	       break;
	  } else if (strcmp(argv[i] + 1, "os-locks") == 0) {
	       USE_SHARED_RO(lock_type) = OS_LOCKS;
	  } else if (strcmp(argv[i] + 1, "memory-locks") == 0) {
	       USE_SHARED_RO(lock_type) = MEMORY_LOCKS;
	  } else
	       break;
     }
     return i;
}

/***********************************************************\
 * Support for mutexen 
\***********************************************************/

#if SPARC_V9
#define ACQUIRE(lock) Cilk_cas(lock, 0, 1)
#define RELEASE(lock) *lock = 0
#endif

#if SPARC_V8
#define ACQUIRE(lock) Cilk_swap(lock, 1)
#define RELEASE(lock) Cilk_swap(lock, 0)
#endif

/*******************************
 * Memory locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_memory(Cilk_mutex *lock)
{
     lock->memory = 1;
     RELEASE(lock);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_memory(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     /* 
      * This hackery should reduce bus traffic, if the
      * cache is snooped correctly.
      */
     while (ACQUIRE(&lock->memory) != 0) {
	  while (lock->memory != 0);
     }

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_memory(Cilk_mutex *lock)
{
     RELEASE(&lock->memory);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_memory(Cilk_mutex *lock)
{
     if (ACQUIRE(&lock->memory) == 0) {
	  return 1;
     } else
	  return 0;
}

/*******************************
 * Operating System locks
 *******************************/
/* initialize lock */
static inline void Cilk_mutex_init_os(Cilk_mutex *lock)
{
     mutex_init(&lock->os, USYNC_PROCESS, NULL);
}

/*
 *  Obtain lock <lock>.  Waits until lock is available.
 */
static inline void Cilk_mutex_wait_os(Cilk_mutex *lock)
{
     Cilk_enter_state(STATE_WAITING_FOR_LOCK);

     mutex_lock(&lock->os);

     Cilk_exit_state(STATE_WAITING_FOR_LOCK);
}

/*
 *  Give up lock <lock>.
 */
static inline void Cilk_mutex_signal_os(Cilk_mutex *lock)
{
     mutex_unlock(&lock->os);
}

/*
 *  Try to obtain lock <lock>.  Returns 1 if it succeeds and 0 if it fails.
 *  Does not wait.
 */
static inline int Cilk_mutex_try_os(Cilk_mutex *lock)
{
     if (mutex_trylock(&lock->os) == 0) {
	  return 1;
     } else {
	  return 0;
     }
}

/*********************
 * Lock glue 
 *********************/
#define MAKE_LOCKFUN(name) \
void name(Cilk_mutex *lock)                                   \
{                                                             \
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {	      \
	  name##_memory(lock);				      \
     } else {						      \
	  name##_os(lock);				      \
     }							      \
}							      \
							       
MAKE_LOCKFUN(Cilk_mutex_init)
MAKE_LOCKFUN(Cilk_mutex_wait)
MAKE_LOCKFUN(Cilk_mutex_signal)

int Cilk_mutex_try(Cilk_mutex *lock)
{
     if (USE_SHARED_RO(lock_type) == MEMORY_LOCKS) {
	  return Cilk_mutex_try_memory(lock);
     } else {
	  return Cilk_mutex_try_os(lock);
     }
}

/***********************************************************\
 * Spawning worker processes
\***********************************************************/

#include <sys/wait.h>
#include <malloc.h>
#include <signal.h>
#include <sys/resource.h>

/*
 * Start forking processes: one process enters this function,
 * <Cilk_active_size> processes should exit this function immediately
 * and return 0.  Each returning process should set <Self> to a unique
 * number between 0 and <Cilk_active_size> - 1.  
 */
void Cilk_create_children(void (*child)(int))
{
     int i, n;
     pid_t pid;
     int stat;
     pid_t *children_pid;


     children_pid = (pid_t *) malloc(USE_SHARED_RO(active_size) 
				     * sizeof(pid_t));
     CILK_CHECK(children_pid,
		("couldn't malloc() children_pid\n"));

     for (i = 0; i < USE_SHARED_RO(active_size); i++) {
	  if ((pid = fork()) < 0) {
	       Cilk_die("can't fork");
	  } else if (pid == 0) {	/* child */
	       child(i);
	       exit(0);
	  } else {		/* parent */
	       children_pid[i] = pid;
	  }
     }

     /* parent now waits for all children to complete */
     for (n = 0; n < USE_SHARED_RO(active_size); n++) {
	  pid = waitpid((pid_t) - 1, &stat, 0);

	  /* find completed child's # */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       if (pid == children_pid[i])
		    break;
	  CILK_ASSERT(i < USE_SHARED_RO(active_size));

	  if (WIFEXITED(stat)) {
	       if (WEXITSTATUS(stat) == 0) {
		    continue;
	       } else
		    Cilk_dprintf("%d: Child exited with non-zero value %d\n",
				 i, WEXITSTATUS(stat));
	  } else if (WIFSIGNALED(stat)) {
	       const char *core;
	       if (WCOREDUMP(stat)) {
		    core = " (core dumped)";
		    /* prevent others from overwriting core, if possible */
		    Cilk_mutex_try(&USE_SHARED(die_lock));
	       } else
		    core = "";

	       /* 
	        * Print out understandable messages for common signals,
	        * and the signal number for the uncommon ones. 
	        */
	       switch (WTERMSIG(stat)) {
		   case SIGILL:
			Cilk_dprintf("%d: Illegal instruction%s\n", i, core);
			break;
		   case SIGABRT:
			Cilk_dprintf("%d: Abort or Assertion failed%s\n", 
				     i, core);
			break;
		   case SIGFPE:
			Cilk_dprintf("%d: Floating exception%s\n", i, core);
			break;
		   case SIGKILL:
			Cilk_dprintf("%d: Terminated%s\n", i, core);
			break;
		   case SIGBUS:
			Cilk_dprintf("%d: Bus error%s\n", i, core);
			break;
		   case SIGSEGV:
			Cilk_dprintf("%d: Segmentation fault%s\n", i, core);
			break;
		   default:
			Cilk_dprintf(
			  "%d: signal %d (see /usr/include/sys/signal.h)"
					 "%s\n", i, WTERMSIG(stat), core);
			break;
	       }
	  }
	  /* 
	   * something went wrong with child i: abort the rest of the 
	   * children.  Sending a SIGKILL keeps them from dumping core.
	   */
	  for (i = 0; i < USE_SHARED_RO(active_size); i++)
	       kill(children_pid[i], SIGKILL);
     }
}

/***************************************************
 * Timer control
 ***************************************************/
static int openslashproc(int pid)
{
     char buf[64];
     int  fd;

     sprintf(buf, "/proc/%d", pid);
     if ( (fd = open(buf, 2)) < 0 )
	  perror("open /proc");

     return fd;
}

static void start_microaccounting(void)
{
     int fd, flags;

     fd = openslashproc(getpid());

     flags = PR_MSACCT;
     if (ioctl(fd, PIOCSET, &flags, sizeof (flags)) < 0) {
	  perror("ioctl set microstate acct");
     }
}


/**********************************************************
 * Priority control
 **********************************************************/
#define LOW_PRIORITY (-60)

DECLARE_PRIVATE_VARIABLE(static int, current_priority,);
void Cilk_lower_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) != LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_PID, getpid(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = tsparams->ts_upri = LOW_PRIORITY;
	  priocntl(P_PID, getpid(), PC_SETPARMS, (caddr_t) & params);
     }
}

void Cilk_raise_priority(void)
{
     struct pcparms params;
     struct tsparms *tsparams = (struct tsparms *) &params.pc_clparms;

     if (USE_PRIVATE(current_priority) == LOW_PRIORITY) {
	  params.pc_cid = PC_CLNULL;
	  priocntl(P_PID, getpid(), PC_GETPARMS, (caddr_t) & params);
	  USE_PRIVATE(current_priority) = 
	       tsparams->ts_upri = tsparams->ts_uprilim;
	  priocntl(P_PID, getpid(), PC_SETPARMS, (caddr_t) & params);
     }
}

/**********************************************************
 * Initialization of the module
 **********************************************************/

void Cilk_arch_specific_init(void)
{
     Cilk_mutex_init(&USE_SHARED(dprintf_lock));
     Cilk_mutex_init(&USE_SHARED(die_lock));
     start_microaccounting();
}

void Cilk_arch_specific_per_process_init(void)
{
     USE_PRIVATE(current_priority) = -1;
     start_microaccounting();
}

