/*
 * $Id: cilk.c,v 1.1.1.1.2.2 1997/03/04 03:11:04 fengmd Exp $
 */

/*
 * $Log: cilk.c,v $
 * Revision 1.1.1.1.2.2  1997/03/04 03:11:04  fengmd
 * Merge the change of the Cilk-5 base.
 *
 * Revision 1.2  1997/03/01 17:42:44  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <stdio.h>

/* Hack: main is defined to be cilk_main; undef it */
#undef main

DECLARE_SHARED_RO_VARIABLE(static Cilk_time, start_time, );

/* globally-visible options (read-only in child processes) */
DECLARE_SHARED_RO_VARIABLE(int, active_size, );
DECLARE_SHARED_RO_VARIABLE(int, StatLevel, );
DECLARE_SHARED_RO_VARIABLE(int, yield_slice, );
DECLARE_SHARED_RO_VARIABLE(size_t, stackdepth, );
DECLARE_SHARED_RO_VARIABLE(char *, infofile_name, );
			/* file where info messages are sent */

/*
 * HACK: this should be FILE *; but using a void * simplifies
 * the automatic generation of the file containing all
 * shared/private/ro variables.
 */
DECLARE_SHARED_RO_VARIABLE(void *, infofile, );

/*
 * variables that children need to update (typically at the end
 * of the program
 */
DECLARE_SHARED_VARIABLE(Cilk_time, critical_path,);
DECLARE_SHARED_VARIABLE(Cilk_time, total_work, );
DECLARE_SHARED_RO_VARIABLE(unsigned int, num_threads, );
DECLARE_SHARED_RO_VARIABLE(unsigned int, num_steals, );

/*
 * this string is printed when an assertion fails.
 * If we just inline it, apparently gcc generates many copies
 * of the string.
 */
DECLARE_SHARED_RO_VARIABLE(const char *, assertion_failed_msg, );

static char *smart_sprint_time(double x)
{
     static char buf[128];

     if (x < 1.0E-3)
	  sprintf(buf, "%f us", x * 1.0E6);
     else if (x < 1.0)
	  sprintf(buf, "%f ms", x * 1.0E3);
     else
	  sprintf(buf, "%f s", x);

     return buf;
}

/*
 * The statistics printer 
 */
static void print_all_statistics(void)
{
     if (strcmp(USE_SHARED_RO(infofile_name), "-") == 0)
	  USE_SHARED_RO(infofile) = stderr;
     else
	  USE_SHARED_RO(infofile) = 
	       fopen(USE_SHARED_RO(infofile_name), "w");

     CILK_CHECK(USE_SHARED_RO(infofile),
		("can't open %s for writing\n", 
		 USE_SHARED_RO(infofile_name)));

     if (USE_SHARED_RO(StatLevel) >= 1) {
	  /* level 1 and above: print wall clock, work and CP */
	  fprintf(USE_SHARED_RO(infofile),
		  "Wall-clock running time on %d processor%s: %s\n",
		  USE_SHARED_RO(active_size), 
		  USE_SHARED_RO(active_size) > 1 ? "s" : "",
		  smart_sprint_time(Cilk_wall_time_to_sec(
		       Cilk_get_wall_time() - USE_SHARED_RO(start_time))));
	  WHEN_CILK_TIMING({
	       fprintf(USE_SHARED_RO(infofile), "Total work = %s\n",
		       smart_sprint_time(Cilk_compute_work()));
	  });

	  /* 
	   * hack: we don't consider the total work if it
	   * is too small.  The trick allows us to compile
	   * the rts with CILK_CRITICAL_PATH=something and the
	   * user program with CILK_CRITICAL_PATH=0
	   */
	  WHEN_CILK_CRITICAL_PATH({
	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001) {
		    fprintf(USE_SHARED_RO(infofile),
			    "Total work (accumulated) = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_SHARED(total_work))));
		    fprintf(USE_SHARED_RO(infofile),
			    "Critical path = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_SHARED(critical_path))));
		    fprintf(USE_SHARED_RO(infofile),
			    "Average parallelism = %f\n",
			    (double) USE_SHARED(total_work) /
			    (double) USE_SHARED(critical_path));
	       }
	  });
     }

     if (USE_SHARED_RO(StatLevel) >= 2) {
	  /* 
	   * level 2 and above: print statistics on work
	   * stealing etc 
	   */
	  Cilk_print_rts_statistics();
	  WHEN_CILK_TIMING({
	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001
		   && USE_SHARED_RO(num_threads) > 0) {
		    fprintf(USE_SHARED_RO(infofile), 
			    "AVERAGE THREAD LENGTH = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(
				      USE_SHARED(total_work)) /
				 (double) USE_SHARED_RO(num_threads)));
	       }

	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001
		   && USE_SHARED_RO(num_steals) > 0) {
		    fprintf(USE_SHARED_RO(infofile),
			    "AVERAGE SUBCOMPUTATION LENGTH = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(
				      USE_SHARED(total_work)) /
				 ((double) USE_SHARED_RO(num_steals) + 1.0)));
	       }
	  });
     }
     if (USE_SHARED_RO(StatLevel) >= 3) {
	  /* level 3 and above: memory statistics */
	  Cilk_internal_malloc_print_statistics();
     }
     if (USE_SHARED_RO(StatLevel) >= 4) {
	  /* level 4 and above: print work done by each processor */
	  Cilk_print_time_statistics();
     }
     if (USE_SHARED_RO(StatLevel) >= 5) {
	  /* level 5 and above: print summary breakout of time */
	  Cilk_summarize_time_statistics();
     }
     if (USE_SHARED_RO(StatLevel) >= 6) {
	  /* level 6 and above: print detailed breakout of time */
	  Cilk_print_detailed_time_statistics();
     }

     fclose(USE_SHARED_RO(infofile));
}

#if CILK_MANAGE_SHARED_SEGMENT
/*
 * If linker segments are used, a machine-specific version
 * must be provided in the architecture-specific file
 */

struct Cilk_shared_variables *Cilk_shared_vars;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_map_shared_segment(void)
{
     Cilk_shared_vars = 
	  Cilk_malloc_fixed(sizeof (struct Cilk_shared_variables));
     CILK_CHECK(Cilk_shared_vars,
		("Can't allocate shared variables!\n"));
}
#endif

#if CILK_MANAGE_PRIVATE_SEGMENT
/*
 * this is really useless, since if Cilk_private_vars
 * is really private, then all the other variables can
 * be private as well.  But at least the mechanism is
 * here, so that if Cilk_private_vars can be put in
 * some register or special location, everything works.
 */
CILK_DEFINE_REALLY_PRIVATE(struct Cilk_private_variables *, Cilk_private_vars);

static struct Cilk_private_variables dummy_private;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_map_private_segment(void)
{
     /* 
      * we use a dummy value because malloc() collects some statistics
      * that use private variables --- aaargh
      */
     Cilk_private_vars = &dummy_private;
     Self = -1;
     Cilk_private_vars = 
	  Cilk_malloc_fixed(sizeof (struct Cilk_private_variables));
     CILK_CHECK(Cilk_private_vars,
		("Can't allocate shared variables!\n"));
}

#endif

static void init_variables(Options *options)
{
     /* initialize all shared_ro variables declared above */
     INIT_SHARED_RO(start_time, Cilk_get_wall_time());
     INIT_SHARED_RO(active_size, options->nproc);
     INIT_SHARED_RO(StatLevel, options->statlevel);
     INIT_SHARED_RO(yield_slice, options->yieldslice);
     INIT_SHARED_RO(stackdepth, options->stackdepth);
     INIT_SHARED_RO(infofile_name, options->infofile_name);
     INIT_SHARED_RO(infofile, (FILE *)0);
     INIT_SHARED_RO(num_threads, 0);
     INIT_SHARED_RO(num_steals, 0);
     INIT_SHARED_RO(assertion_failed_msg,
		    "Assertion failed: %s line %d file %s\n"
		    "Ex absurdo sequitur quodlibet.\n");
}

/* global init */
static void global_init(Options *options)
{
     init_variables(options);

     /*
      * set up malloc() arena; this must be done first
      */
     Cilk_init_arena(options->heap);

     /* setup the address space */
     Cilk_map_shared_segment();
     Cilk_map_private_segment();

     /* set weird self value */
     Self = -1;
     USE_SHARED(critical_path) = 0;
     USE_SHARED(total_work) = 0;

     /* architecture-specific module */
     Cilk_arch_specific_init();

     /* initialize barrier code */
     Cilk_barrier_init();

     /* initialize the statistics gatherer and timer */
     Cilk_event_gathering_init();
     Cilk_time_gathering_init();

     /* run initialization hooks for library functions */
     Cilk_run_hooks(Cilk_init_global_hooks);
}

DECLARE_SHARED_RO_VARIABLE(static struct Closure *, invoke_main, );

/*
 * this function is the entry point for child processes
 */
static void child_main(int id)
{
     Cilk_map_private_segment();
     Self = id;

     Cilk_arch_specific_per_process_init();

     /* run initialization hooks for library functions */
     Cilk_run_hooks(Cilk_init_per_process_hooks);

     WHEN_CILK_ND(nondeterminator_init());

     Cilk_barrier();
     
     if (Self == 0)
	  Cilk_scheduler(USE_SHARED_RO(invoke_main));
     else
	  Cilk_scheduler((Closure *) NULL);
}


int main(int argc, char *argv[])
{
     Options options;

     /* parse command-line arguments */
     if (Cilk_parse_command_line(&options, &argc, argv))
	  return 0;

     /* initialize global part of system */
     global_init(&options);

     /* initialize the scheduler */
     Cilk_scheduler_init();
     USE_SHARED_RO(invoke_main) = Cilk_create_initial_thread(argc, argv);

     /* create children; when this returns we are done */
     Cilk_create_children(child_main);

     Cilk_internal_malloc_final_check();

     /* everybody is done: print stats and exit */
     print_all_statistics();
     exit(USE_SHARED(retval));
     return -1;
}

extern void Cilk_exit_from_user_main(Closure *cl, int res);

void Cilk_exit(int res){

     Cilk_exit_from_user_main(USE_SHARED_RO(invoke_main), res);

}
















