/*
 * $Id: cmdline.c,v 1.1.1.1.2.4 1997/03/04 03:11:07 fengmd Exp $
 */

/*
 * $Log: cmdline.c,v $
 * Revision 1.1.1.1.2.4  1997/03/04 03:11:07  fengmd
 * Merge the change of the Cilk-5 base.
 *
 * Revision 1.2  1997/03/01 17:42:49  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Command line parsing
\***********************************************************/

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <stdio.h>

enum {
     NONE, NPROC, STATS, NO_STATS, HELP, ARENA, STACK, YIELD, NO_YIELD,
     INFOFILE, FIRST_STOP, FLOAT_COMM, ND_ARENA, END_OPTIONS
};

static struct options {
     char *string;
     int option;
     char *help;
} optarray[] = {
     {
	  "-", END_OPTIONS, "-- : end of option parsing"
     },
     {
	  "nproc", NPROC, "-nproc <n> : set number of processors"
     },
     {
	  "stack", STACK, "-stack <n> : set stack depth"
     },
     {
	  "stats", STATS, "-stats <n> : set statistics level"
     },
     {
	  "no-stats", NO_STATS, "-no-stats : disable statistics printing"
     },
     {
	  "help", HELP, "-help : print this message"
     },
     {
	  "heap", ARENA, "-heap <n> : set heap size (in Mb)"
     },
     {
	  "yield", YIELD, "-yield : yield time slice when steal fails"
     },
     {
	  "no-yield", NO_YIELD, "-no-yield : don't yield time slice"
     },
     {
	  "info", INFOFILE, "-info <filename> : set name of info file"
     },
     {
	  "infofile", INFOFILE, "-infofile : same as -info"
     },
#ifdef CILK_ND
     {
          "first-stop", FIRST_STOP, "-first-stop : stop the program when the first race is found"
     },
     {
          "float-comm", FLOAT_COMM, "-float-comm : assuming floating point operations are commutative"
     },
     {
          "nd-heap", ND_ARENA, "-nd-heap <n> : set the Nondeterminator heap size (in Mb)"
     },
#endif
     {
	  (char *) 0, NONE, ""
     }
};

static void print_help(void)
{
     struct options *p;

     fprintf(stderr, "Cilk options:\n");

     for (p = optarray; p->string; ++p)
	  if (p->help)
	       fprintf(stderr, "     %s\n", p->help);

     fprintf(stderr, "\n");
}

/* look for a given string in the option table */
static struct options *parse_option(char *s)
{
     struct options *p;

     for (p = optarray; p->string; ++p)
	  if (strcmp(s, p->string) == 0)
	       break;

     return p;
}

#define CHECK(cond, complaint) \
     if (!(cond)) { fprintf(stderr, "Bad option argument for -%s: %s\n", p->string, complaint); return 1; }

int Cilk_parse_command_line(Options *options, int *argc, char *argv[])
{
     int i, j;
     struct options *p;

     /* default options */
     *options = DEFAULT_OPTIONS;

     WHEN_CILK_ND(INIT_SHARED_RO(nd_arena, DEFAULT_ND_HEAPSIZE);)

     for (i = 1; i < *argc; ++i) {
	  if (*argv[i] != '-')
	       goto arch_specific;
	  p = parse_option(argv[i] + 1);
	  switch (p->option) {
	      case NPROC:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   options->nproc = atoi(argv[i]);
		   if (options->nproc == 0)
			options->nproc = Cilk_partition_size();
		   CHECK(options->nproc > 0,
			 "non-positive number of processors");
		   break;
	      case STACK:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   options->stackdepth = atoi(argv[i]);
		   CHECK(options->stackdepth > 0,
			 "non-positive stack depth");
		   break;
	      case STATS:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   options->statlevel = atoi(argv[i]);
		   break;
	      case NO_STATS:
		   options->statlevel = 0;
		   break;
	      case YIELD:
		   options->yieldslice = 1;
		   break;
	      case NO_YIELD:
		   options->yieldslice = 0;
		   break;
	      case HELP:
		   print_help();
		   return 1;
		   break;
	      case ARENA:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   options->heap = atof(argv[i]) * MEGABYTE;
		   CHECK(options->heap > 0,
			 "non-positive heap size");
		   break;
	      case INFOFILE:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   options->infofile_name = argv[i];
		   break;
	      WHEN_CILK_ND(
	      case FIRST_STOP:
	           INIT_SHARED_RO(first_stop, 1);
		   break;
	      case FLOAT_COMM:
	           INIT_SHARED_RO(float_comm, 1);
		   break;
	      case ND_ARENA:
		   ++i;
		   CHECK(i < *argc,
			 "argument missing");
		   INIT_SHARED_RO(nd_arena, atof(argv[i]) * MEGABYTE);
		   CHECK(USE_SHARED_RO(nd_arena) > 0,
			 "non-positive Nondeterminator heap size");
		   break;
              )
	      case END_OPTIONS:
		   ++i;
		   goto stop;
	      default:
		   goto arch_specific;
	  }
     }

     /* let architecture-specific code get a crack at the command line */
   arch_specific:
     i = Cilk_arch_specific_options(i, *argc, argv);

     /* remove Cilk command-line arguments from argv */
   stop:
     for (j = i; j <= *argc; j++)
	  argv[j - (i - 1)] = argv[j];
     *argc -= i - 1;

     return 0;
}
