/*
 * $Id: malloc.c,v 1.1.1.1.2.4 1997/03/04 03:11:14 fengmd Exp $
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * This is a simple malloc()/free() for Cilk. I've looked at
 * GNU and BSD malloc(), and they do dirty things. Since hacking
 * is fun, I'll do dirty things myself. This malloc() is in the
 * spirit of BSD, rather than GNU.
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#if CILK_IRIX
#include <bstring.h>
#else
#include <strings.h>
#endif

/*
 * local variables used by malloc
 */
struct malloc_info {
     Cilk_mutex malloc_lock;
     char *arena, *end_arena;
     struct overhead *freelist[CILK_MALLOC_NLISTS];
};

DECLARE_SHARED_RO_VARIABLE(static struct malloc_info *, malloc_info, );

#define malloc_lock (USE_SHARED_RO(malloc_info)->malloc_lock)
#define arena (USE_SHARED_RO(malloc_info)->arena)
#define end_arena (USE_SHARED_RO(malloc_info)->end_arena)
#define freelist (USE_SHARED_RO(malloc_info)->freelist)

/*
 * General idea: we have the amount n of allocated memory
 * is rounded to a small set of integers (currently powers of
 * two, but the trick is more general). There is a free list
 * for each rounded size. Then each free list is managed separately
 * in the obvious way. A nuisance is that malloc() needs to keep
 * track of the size of the block.
 */

/* 
 * this struct contains
 * 1) pointer to free list, if block is allocated.
 * 2) pointer to next free block, if block is free.
 */
struct overhead {
     void *next;
     struct overhead *actual;	/* 
				 * where this structure would be, in absence
				 * of alignment constraints
				 */
     WHEN_CILK_DEBUG(int magic;)
     /* keep also a magic number for debugging */
};

#define ALLOC_MAGIC 0x4213BDEF
#define FREE_MAGIC 0x42BDEF12
#define FIXED_MAGIC 0x42321BDF

/* must be 2^k */
#define ALIGNMENT CILK_MAX_ALIGN /* 
				 * every malloc() object has this 
				 * alignment  
				 */

#define ALIGN(s) ((typeof(s)) (((size_t) s + (ALIGNMENT - 1)) & (~(ALIGNMENT - 1))))

#define EXTRA_SPACE 64		/* 
				 * 64 is a magic number 
				 * and may double performance of
				 * certain programs :-)
				 */
#define MINSIZE 8

DECLARE_SHARED_RO_VARIABLE(static int, pagesize, );

/* 
 * allocate an "arena" of shared memory of the given size.
 * This is the area of memory from which malloc() gets its
 * blocks.
 */
void Cilk_init_arena(size_t arenasize)
{
     int i;
     size_t size;

     CILK_ASSERT((ALIGNMENT & (ALIGNMENT-1)) == 0);

     INIT_SHARED_RO(pagesize, Cilk_get_page_size());
     size = (arenasize + USE_SHARED_RO(pagesize) - 1) / 
	  USE_SHARED_RO(pagesize) * USE_SHARED_RO(pagesize);

     /*
      * begin of arena is the malloc description block,
      * containing, in particular, the lock and the arena
      * pointer 
      */
     USE_SHARED_RO(malloc_info) = Cilk_map_shared(size);
     arena = ((char *)USE_SHARED_RO(malloc_info)) + 
	  sizeof(struct malloc_info);
     end_arena = ((char *)USE_SHARED_RO(malloc_info)) + size;

     WHEN_CILK_ND(INIT_SHARED_RO(heap_arena_start,
				 (char *)USE_SHARED_RO(malloc_info));)
     WHEN_CILK_ND(INIT_SHARED_RO(heap_arena_end, end_arena);)

     arena = ALIGN(arena);

     /* now initialize all free lists */
     for (i = 0; i < CILK_MALLOC_NLISTS; ++i)
	  freelist[i] = (struct overhead *) 0;

     /* init the lock */
     Cilk_mutex_init(&malloc_lock);
}

/* allocate some chunk of the arena */
static void *morecore(size_t size)
{
     char *old_arena = arena;

     CILK_CHECK(size < end_arena - arena,
		("Cilk malloc() memory exhausted, can't alloc %d bytes\n",
		 size));

     arena += size;

     arena = ALIGN(arena);

     return old_arena;
}

/* memory is never returned to the arena in this implementation */

/*
 * get a pointer to the free list corresponding to size.
 * First of all, round size.
 */
static struct overhead **getbucket(size_t size)
{
     unsigned long long k;
     struct overhead **p;

     /* minimum size of 8 bytes */
     k = MINSIZE;

     /* find next power of two, and appropriate pointer */
     for (p = freelist; 
	  k + EXTRA_SPACE < (unsigned long long) size; ++p)
	  k <<= 1;

     /* 
      * now here are two cases: either the free list is empty
      * or not.
      */
     if (*p == NULL) {
	  /* empty free list. Get block from arena */
	  *p = morecore(k + EXTRA_SPACE);

	  (*p)->next = NULL;

	  /* declare it free */
	  WHEN_CILK_DEBUG((*p)->magic = FREE_MAGIC);
     }
     return p;
}

/*
 * now Cilk_malloc() is easy.
 */
void *Cilk_malloc(size_t size)
{
     struct overhead **p;
     struct overhead *q;

     /* what about this? Should we return NULL ? malloc() sucks */
#if BRAINDEAD
     if (size == 0)
	  size = 1;
#endif

     CILK_CHECK(size > 0,
		("Wrong argument %d to Cilk_malloc()\n", size));

     /* first of all, get the lock */
     Cilk_mutex_wait(&malloc_lock);

     /* get a bucket, also counting overhead */
     p = getbucket(size + ALIGN(sizeof(struct overhead)));

     /* (*p) != NULL, but be paranoid anyway */
     CILK_ASSERT(*p);
     CILK_ASSERT((*p)->magic == FREE_MAGIC);

     q = *p;
     CILK_ASSERT(q == ALIGN(q));

     /* remove elem from free list */
     *p = q->next;

     /* record the free list in q */
     q->next = p;
     q->actual = q;
     CILK_ASSERT((q->magic = ALLOC_MAGIC) || 1);

     /* give up the lock */
     Cilk_mutex_signal(&malloc_lock);

     return (((char *) q) + ALIGN(sizeof(struct overhead)));
}

/*
 * Cilk_malloc_fixed allocates something that will not be freed
 * The advantage is that the size is not rounded up.
 */
void *Cilk_malloc_fixed(size_t size)
{
     void *p;
     CILK_CHECK(size > 0,
		("Wrong argument %d to Cilk_malloc()\n", size));

     /* first of all, get the lock */
     Cilk_mutex_wait(&malloc_lock);

     size = ALIGN(size);
     p = morecore(size);

     /* give up the lock */
     Cilk_mutex_signal(&malloc_lock);
     return p;
}

/*
 * Cilk_memalign works as follows: a big enough block 
 * is allocated using malloc, and the correctly aligned address
 * is computed. Then the overhead info is relocated to the
 * new address, and the new address returned.
 * Cilk_free() brings back the overhead info to the right place.
 */
void *Cilk_memalign(size_t alignment, size_t size)
{
     void *q;
     struct overhead o;

     CILK_CHECK(size > 0 && alignment > 0,
		("Wrong arguments to Cilk_memalign(%d, %d)\n",
		 alignment, size));

     /* 
      * first of all, ensure that alignment satisfies also
      * the internal constraints.
      * A better policy is probably to disallow non-power-of-2
      * alignments.
      */
     while (alignment != ALIGN(alignment))
	  alignment <<= 1;

     /* adjust size to be big enough to allow relocation */
     size = size + alignment - 1;

     /* now use malloc to get a block */
     q = Cilk_malloc(size);

     /* get old overhead */
     o = *(struct overhead *)
	 (((char *) q) - ALIGN(sizeof(struct overhead)));

     /* align q */
     q = (void *)
	 ((((unsigned int) q + alignment - 1) / alignment) * alignment);

     /* relocate overhead */
     *(struct overhead *)
	 (((char *) q) - ALIGN(sizeof(struct overhead))) = o;

     return q;
}

/*
 * Cilk_free() is easy, too 
 */
void Cilk_free(void *s)
{
     struct overhead *q = (struct overhead *)
     (((char *) s) - ALIGN(sizeof(struct overhead)));
     struct overhead **p;

     CILK_ASSERT(q == ALIGN(q));

     /* first of all, get the lock */
     Cilk_mutex_wait(&malloc_lock);

     CILK_ASSERT(q->magic == ALLOC_MAGIC);
     p = q->next;

     CILK_ASSERT(p >= freelist && 
		 p < freelist + CILK_MALLOC_NLISTS);

     WHEN_CILK_ND( {
       /* clean up shadow space */
       struct overhead **pp;
       int k = MINSIZE;

       for (pp = freelist; pp < p; ++pp, k<<= 1);
       CILK_ASSERT(pp == p);
       k += EXTRA_SPACE - ALIGN(sizeof(struct overhead));
       memset(HEAP_READER(s), S_SET, k);
       memset(HEAP_WRITER(s), S_SET, k);
     } )

     /* restore q to its actual value */
     q = q->actual;
     CILK_ASSERT((q->magic = FREE_MAGIC) || 1);

     /* q goes to free list */
     q->next = *p;
     *p = q;

     CILK_ASSERT(*p);

     Cilk_mutex_signal(&malloc_lock);
     return;
}

/*
 * Some other functions, for completeness
 */
void *Cilk_calloc(size_t nelem, size_t elsize)
{
     int size = nelem * elsize;
     char *mem = Cilk_malloc(size);
     if (mem != NULL)
	  bzero(mem, size);
     return mem;
}

/*
 * This realloc is not very efficient, but it works.  Also, this
 * realloc will not take a pointer that has already been freed.
 * (see SunOS 4.x man pages for realloc for why this is a problem)
 */
void *Cilk_realloc(void *s, size_t size)
{
     if (s == NULL)
	  return Cilk_malloc(size);
     else if (size == 0) {
	  Cilk_free(s);
	  return Cilk_malloc(0);
     } else {
	  struct overhead *q = (struct overhead *)
	  (((char *) s) - ALIGN(sizeof(struct overhead)));
	  struct overhead **p;
	  int oldsize;
	  void *new;

	  CILK_ASSERT(q == ALIGN(q));
	  CILK_ASSERT(q->magic == ALLOC_MAGIC);

	  p = q->next;
	  CILK_ASSERT(p >= freelist &&
		      p < freelist + CILK_MALLOC_NLISTS);

	  /* compute size of s */
	  oldsize = MINSIZE;
	  while (p > freelist) {
	       p--;
	       oldsize <<= 1;
	  }

	  oldsize += EXTRA_SPACE;
	  oldsize -= ALIGN(sizeof(struct overhead));

	  new = Cilk_malloc(size);
	  bcopy(s, new, oldsize < size ? oldsize : size);

	  Cilk_free(s);

	  return new;
     }
}

void *Cilk_valloc(size_t size)
{
     return Cilk_memalign(USE_SHARED_RO(pagesize), size);
}
